@@ -1,7 +1,8 @@
 /* -*- c-basic-offset: 8; -*- */
 /* shout.c: Implementation of public libshout interface shout.h
  *
- *  Copyright (C) 2002-2004 the Icecast team <team@icecast.org>
+ *  Copyright (C) 2002-2004 the Icecast team <team@icecast.org>,
+ *  Copyright (C) 2012-2015 Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -17,7 +18,7 @@
  *  License along with this library; if not, write to the Free
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * $Id: shout.c 18310 2012-05-24 20:25:30Z giles $
+ * $Id$
  */
 
 #ifdef HAVE_CONFIG_H
@@ -28,12 +29,13 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
 #include <errno.h>
 
 #include <shout/shout.h>
-#include <net/sock.h>
-#include "timing/timing.h"
-#include "httpp/httpp.h"
+#include <common/net/sock.h>
+#include "common/timing/timing.h"
+#include "common/httpp/httpp.h"
 
 #include "shout_private.h"
 #include "util.h"
@@ -47,24 +49,13 @@
 #endif
 
 /* -- local prototypes -- */
-static int queue_data(shout_queue_t *queue, const unsigned char *data, size_t len);
-static int queue_str(shout_t *self, const char *str);
-static int queue_printf(shout_t *self, const char *fmt, ...);
-static inline void queue_free(shout_queue_t *queue);
 static int send_queue(shout_t *self);
 static int get_response(shout_t *self);
 static int try_connect (shout_t *self);
 static int try_write (shout_t *self, const void *data, size_t len);
 
 static int create_request(shout_t *self);
-static int create_http_request(shout_t *self);
-static int create_xaudiocast_request(shout_t *self);
-static int create_icy_request(shout_t *self);
 static int parse_response(shout_t *self);
-static int parse_http_response(shout_t *self);
-static int parse_xaudiocast_response(shout_t *self);
-
-static char *http_basic_authorization(shout_t *self);
 
 /* -- static data -- */
 static int _initialized = 0;
@@ -121,6 +112,25 @@ shout_t *shout_new(void)
 
 		return NULL;
 	}
+	if (!(self->meta = _shout_util_dict_new())) {
+		shout_free(self);
+
+		return NULL;
+	}
+	if (shout_set_meta(self, "name", "no name") != SHOUTERR_SUCCESS) {
+		shout_free(self);
+
+		return NULL;
+	}
+
+#ifdef HAVE_OPENSSL
+	if (shout_set_allowed_ciphers(self, LIBSHOUT_DEFAULT_ALLOWED_CIPHERS) != SHOUTERR_SUCCESS) {
+		shout_free(self);
+
+		return NULL;
+	}
+        self->tls_mode = SHOUT_TLS_AUTO;
+#endif
 
 	self->port = LIBSHOUT_DEFAULT_PORT;
 	self->format = LIBSHOUT_DEFAULT_FORMAT;
@@ -133,16 +143,22 @@ void shout_free(shout_t *self)
 {
 	if (!self) return;
 
+        if (self->state != SHOUT_STATE_UNCONNECTED) return;
+
 	if (self->host) free(self->host);
 	if (self->password) free(self->password);
 	if (self->mount) free(self->mount);
-	if (self->name) free(self->name);
-	if (self->url) free(self->url);
-	if (self->genre) free(self->genre);
-	if (self->description) free(self->description);
 	if (self->user) free(self->user);
 	if (self->useragent) free(self->useragent);
 	if (self->audio_info) _shout_util_dict_free (self->audio_info);
+	if (self->meta) _shout_util_dict_free (self->meta);
+
+#ifdef HAVE_OPENSSL
+	if (self->ca_directory) free(self->ca_directory);
+	if (self->ca_file) free(self->ca_file);
+	if (self->allowed_ciphers) free(self->allowed_ciphers);
+	if (self->client_certificate) free(self->client_certificate);
+#endif
 
 	free(self);
 }
@@ -156,7 +172,7 @@ int shout_open(shout_t *self)
 		return SHOUTERR_CONNECTED;
 	if (!self->host || !self->password || !self->port)
 		return self->error = SHOUTERR_INSANE;
-	if (self->format == SHOUT_FORMAT_OGG && self->protocol != SHOUT_PROTOCOL_HTTP)
+	if (self->format == SHOUT_FORMAT_OGG && (self->protocol != SHOUT_PROTOCOL_HTTP && self->protocol != SHOUT_PROTOCOL_ROARAUDIO))
 		return self->error = SHOUTERR_UNSUPPORTED;
 
 	return self->error = try_connect(self);
@@ -174,12 +190,18 @@ int shout_close(shout_t *self)
 	if (self->state == SHOUT_STATE_CONNECTED && self->close)
 		self->close(self);
 
+#ifdef HAVE_OPENSSL
+	if (self->tls)
+		shout_tls_close(self->tls);
+	self->tls = NULL;
+#endif
+
 	sock_close(self->socket);
 	self->state = SHOUT_STATE_UNCONNECTED;
 	self->starttime = 0;
 	self->senttime = 0;
-	queue_free(&self->rqueue);
-	queue_free(&self->wqueue);
+	shout_queue_free(&self->rqueue);
+	shout_queue_free(&self->wqueue);
 
 	return self->error = SHOUTERR_SUCCESS;
 }
@@ -217,8 +239,8 @@ ssize_t shout_send_raw(shout_t *self, const unsigned char *data, size_t len)
 	if (len && ! self->wqueue.len) {
 		if ((ret = try_write(self, data, len)) < 0)
 			return self->error;
-		if (ret < len) {
-			self->error = queue_data(&self->wqueue, data + ret, len - ret);
+		if (ret < (ssize_t)len) {
+			self->error = shout_queue_data(&self->wqueue, data + ret, len - ret);
 			if (self->error != SHOUTERR_SUCCESS)
 				return self->error;
 		}
@@ -226,7 +248,7 @@ ssize_t shout_send_raw(shout_t *self, const unsigned char *data, size_t len)
 		return len;
 	}
 
-	self->error = queue_data(&self->wqueue, data, len);
+	self->error = shout_queue_data(&self->wqueue, data, len);
 	if (self->error != SHOUTERR_SUCCESS)
 		return self->error;
 
@@ -299,40 +321,162 @@ int shout_metadata_add(shout_metadata_t *self, const char *name, const char *val
  * TODO: prettier error-handling. */
 int shout_set_metadata(shout_t *self, shout_metadata_t *metadata)
 {
-	sock_t socket;
+	int error;
+	sock_t socket = -1;
 	int rv;
-	char *encvalue;
+	char *encvalue = NULL;
+	const char *request_template;
+	char *request = NULL;
+	size_t request_len;
+	char *auth = NULL;
+#ifdef HAVE_OPENSSL
+	shout_tls_t *tls = NULL;
+#endif
 
 	if (!self || !metadata)
 		return SHOUTERR_INSANE;
 
 	if (!(encvalue = _shout_util_dict_urlencode(metadata, '&')))
-		return SHOUTERR_MALLOC;
+		goto error_malloc;
+
+	switch (self->protocol) {
+	case SHOUT_PROTOCOL_ICY:
+		request_template = "GET /admin.cgi?mode=updinfo&pass=%s&%s HTTP/1.0\r\nUser-Agent: %s (Mozilla compatible)\r\n\r\n";
+		request_len = strlen(request_template) + strlen(self->password) + strlen(encvalue) + strlen(shout_get_agent(self)) + 1;
+		if (!(request = malloc(request_len)))
+			goto error_malloc;
+		snprintf(request, request_len, request_template, self->password, encvalue, shout_get_agent(self));
+	break;
+	case SHOUT_PROTOCOL_HTTP:
+		auth = shout_http_basic_authorization(self);
+
+		request_template = "GET /admin/metadata?mode=updinfo&mount=%s&%s HTTP/1.0\r\nUser-Agent: %s\r\n%s\r\n";
+		request_len = strlen(request_template) + strlen(self->mount) + strlen(encvalue) + strlen(shout_get_agent(self)) + 1;
+		if (auth)
+			request_len += strlen(auth);
+		if (!(request = malloc(request_len)))
+			goto error_malloc;
+		snprintf(request, request_len, request_template, self->mount, encvalue, shout_get_agent(self), auth ? auth : "");
+	break;
+	default:
+		request_template = "GET /admin.cgi?mode=updinfo&pass=%s&mount=%s&%s HTTP/1.0\r\nUser-Agent: %s\r\n\r\n";
+		request_len = strlen(request_template) + strlen(self->password) + strlen(self->mount) + strlen(encvalue) + strlen(shout_get_agent(self)) + 1;
+		if (!(request = malloc(request_len)))
+			goto error_malloc;
+		snprintf(request, request_len, request_template, self->password, self->mount, encvalue, shout_get_agent(self));
+	break;
+	}
+
+	free(encvalue);
+	encvalue = NULL;
+
+	if (auth)
+		free(auth);
+	auth = NULL;
 
 	if ((socket = sock_connect(self->host, self->port)) <= 0)
 		return SHOUTERR_NOCONNECT;
 
-	if (self->protocol == SHOUT_PROTOCOL_ICY)
-		rv = sock_write(socket, "GET /admin.cgi?mode=updinfo&pass=%s&%s HTTP/1.0\r\nUser-Agent: %s (Mozilla compatible)\r\n\r\n",
-		  self->password, encvalue, shout_get_agent(self));
-	else if (self->protocol == SHOUT_PROTOCOL_HTTP) {
-		char *auth = http_basic_authorization(self);
+#ifdef HAVE_OPENSSL
+	switch (self->tls_mode) {
+	case SHOUT_TLS_DISABLED:
+		/* nothing to do */
+	break;
+	case SHOUT_TLS_RFC2817: /* Use TLS via HTTP Upgrade:-header [RFC2817]. */
+		do { /* use a subscope to avoid more function level variables */
+			char upgrade[512];
+			size_t len;
 
-		rv = sock_write(socket, "GET /admin/metadata?mode=updinfo&mount=%s&%s HTTP/1.0\r\nUser-Agent: %s\r\n%s\r\n",
-		  self->mount, encvalue, shout_get_agent(self), auth ? auth : "");
-                free(auth);
-	} else
-		rv = sock_write(socket, "GET /admin.cgi?mode=updinfo&pass=%s&mount=%s&%s HTTP/1.0\r\nUser-Agent: %s\r\n\r\n",
-		  self->password, self->mount, encvalue, shout_get_agent(self));
-	free(encvalue);
-	if (!rv) {
-		sock_close(socket);
-		return SHOUTERR_SOCKET;
+			/* send upgrade request */
+			snprintf(upgrade, sizeof(upgrade),
+				"GET / HTTP/1.1\r\nConnection: Upgrade\r\nUpgrade: TLS/1.0\r\nHost: %s:%i\r\n\r\n",
+				self->host, self->port);
+			upgrade[sizeof(upgrade)-1] = 0;
+			len = strlen(upgrade);
+			if (len == (sizeof(upgrade) - 1))
+				goto error_malloc;
+			rv = sock_write_bytes(socket, upgrade, len);
+			if (len != (size_t)rv)
+				goto error_socket;
+
+			/* read status line */
+			if (!sock_read_line(socket, upgrade, sizeof(upgrade)))
+				goto error_socket;
+			if (strncmp(upgrade, "HTTP/1.1 101 ", 13) != 0)
+				goto error_socket;
+
+			/* read headers */
+			len = 0;
+			do {
+				if (!sock_read_line(socket, upgrade, sizeof(upgrade)))
+					goto error_socket;
+				if (upgrade[0] == 0)
+					break;
+				if (!strncasecmp(upgrade, "Content-Length: ", 16) == 0)
+					len = atoi(upgrade + 16);
+			} while (1);
+
+			/* read body */
+			while (len) {
+				rv = sock_read_bytes(socket, upgrade, len > sizeof(upgrade) ? sizeof(upgrade) : len);
+				if (rv < 1)
+					goto error_socket;
+				len -= rv;
+			}
+		} while (0);
+	/* fall thru */
+	case SHOUT_TLS_RFC2818: /* Use TLS for transport layer like HTTPS [RFC2818] does. */
+		tls = shout_tls_new(self, socket);
+		if (!tls)
+			goto error_malloc;
+		error = shout_tls_try_connect(tls);
+		if (error != SHOUTERR_SUCCESS)
+			goto error;
+	break;
+	default:
+		/* Bad mode or auto detection not completed. */
+		error = SHOUTERR_INSANE;
+		goto error;
+	break;
 	}
+#endif
 
-	sock_close(socket);
+#ifdef HAVE_OPENSSL
+	if (tls) {
+		rv = shout_tls_write(tls, request, strlen(request));
+	} else {
+		rv = sock_write(socket, "%s", request);
+	}
+#else
+	rv = sock_write(socket, "%s", request);
+#endif
 
-	return SHOUTERR_SUCCESS;
+	if (!rv)
+		goto error_socket;
+
+	error = SHOUTERR_SUCCESS;
+	goto error;
+
+error_socket:
+	error = SHOUTERR_SOCKET;
+	goto error;
+error_malloc:
+	error = SHOUTERR_MALLOC;
+	goto error;
+error:
+#ifdef HAVE_OPENSSL
+	if (tls)
+		shout_tls_close(tls);
+#endif
+	if (socket != -1)
+		sock_close(socket);
+	if (encvalue)
+		free(encvalue);
+	if (request)
+		free(request);
+	if (auth)
+		free(auth);
+	return error;
 }
 
 /* getters/setters */
@@ -379,6 +523,12 @@ const char *shout_get_error(shout_t *self)
                 return "Socket is busy";
 	case SHOUTERR_UNSUPPORTED:
 		return "This libshout doesn't support the requested option";
+	case SHOUTERR_NOTLS:
+		return "TLS requested but not supported by peer";
+	case SHOUTERR_TLSBADCERT:
+		return "TLS connection can not be established because of bad certificate";
+	case SHOUTERR_RETRY:
+		return "Please retry current operation.";
 	default:
 		return "Unknown error";
 	}
@@ -498,7 +648,7 @@ int shout_set_mount(shout_t *self, const char *mount)
 	if (!(self->mount = malloc(len)))
 		return self->error = SHOUTERR_MALLOC;
 
-	sprintf (self->mount, "%s%s", mount[0] == '/' ? "" : "/", mount);
+	snprintf (self->mount, len, "%s%s", mount[0] == '/' ? "" : "/", mount);
 
 	return self->error = SHOUTERR_SUCCESS;
 }
@@ -513,77 +663,32 @@ const char *shout_get_mount(shout_t *self)
 
 int shout_set_name(shout_t *self, const char *name)
 {
-	if (!self)
-		return SHOUTERR_INSANE;
-
-	if (self->state != SHOUT_STATE_UNCONNECTED)
-		return self->error = SHOUTERR_CONNECTED;
-
-	if (self->name)
-		free(self->name);
-
-	if (!(self->name = _shout_util_strdup(name)))
-		return self->error = SHOUTERR_MALLOC;
-
-	return self->error = SHOUTERR_SUCCESS;
+	return shout_set_meta(self, "name", name);
 }
 
 const char *shout_get_name(shout_t *self)
 {
-	if (!self)
-		return NULL;
-
-	return self->name;
+	return shout_get_meta(self, "name");
 }
 
 int shout_set_url(shout_t *self, const char *url)
 {
-	if (!self)
-		return SHOUTERR_INSANE;
-
-	if (self->state != SHOUT_STATE_UNCONNECTED)
-		return self->error = SHOUTERR_CONNECTED;
-
-	if (self->url)
-		free(self->url);
-
-	if (!(self->url = _shout_util_strdup(url)))
-		return self->error = SHOUTERR_MALLOC;
-
-	return self->error = SHOUTERR_SUCCESS;
+	return shout_set_meta(self, "url", url);
 }
 
 const char *shout_get_url(shout_t *self)
 {
-	if (!self)
-		return NULL;
-
-	return self->url;
+	return shout_get_meta(self, "url");
 }
 
 int shout_set_genre(shout_t *self, const char *genre)
 {
-	if (!self)
-		return SHOUTERR_INSANE;
-
-	if (self->state != SHOUT_STATE_UNCONNECTED)
-		return self->error = SHOUTERR_CONNECTED;
-
-	if (self->genre)
-		free(self->genre);
-
-	if (! (self->genre = _shout_util_strdup (genre)))
-		return self->error = SHOUTERR_MALLOC;
-
-	return self->error = SHOUTERR_SUCCESS;
+	return shout_set_meta(self, "genre", genre);
 }
 
 const char *shout_get_genre(shout_t *self)
 {
-	if (!self)
-		return NULL;
-
-	return self->genre;
+	return shout_get_meta(self, "genre");
 }
 
 int shout_set_agent(shout_t *self, const char *agent)
@@ -639,27 +744,12 @@ const char *shout_get_user(shout_t *self)
 
 int shout_set_description(shout_t *self, const char *description)
 {
-	if (!self)
-		return SHOUTERR_INSANE;
-
-	if (self->state != SHOUT_STATE_UNCONNECTED)
-		return self->error = SHOUTERR_CONNECTED;
-
-	if (self->description)
-		free(self->description);
-
-	if (! (self->description = _shout_util_strdup (description)))
-		return self->error = SHOUTERR_MALLOC;
-
-	return self->error = SHOUTERR_SUCCESS;
+	return shout_set_meta(self, "description", description);
 }
 
 const char *shout_get_description(shout_t *self)
 {
-	if (!self)
-		return NULL;
-
-	return self->description;
+	return shout_get_meta(self, "description");
 }
 
 int shout_set_dumpfile(shout_t *self, const char *dumpfile)
@@ -689,14 +779,45 @@ const char *shout_get_dumpfile(shout_t *self)
 
 int shout_set_audio_info(shout_t *self, const char *name, const char *value)
 {
+	if (!self)
+		return SHOUTERR_INSANE;
+
 	return self->error = _shout_util_dict_set(self->audio_info, name, value);
 }
 
 const char *shout_get_audio_info(shout_t *self, const char *name)
 {
+	if (!self)
+		return NULL;
+
 	return _shout_util_dict_get(self->audio_info, name);
 }
 
+int shout_set_meta(shout_t *self, const char *name, const char *value)
+{
+	size_t i;
+
+	if (!self || !name)
+		return SHOUTERR_INSANE;
+
+	if (self->state != SHOUT_STATE_UNCONNECTED)
+		return self->error = SHOUTERR_CONNECTED;
+
+	for (i = 0; name[i]; i++)
+		if ((name[i] < 'a' || name[i] > 'z') && (name[i] < '0' || name[i] > '9'))
+			return self->error = SHOUTERR_INSANE;
+
+	return self->error = _shout_util_dict_set(self->meta, name, value);
+}
+
+const char *shout_get_meta(shout_t *self, const char *name)
+{
+	if (!self)
+		return NULL;
+
+	return _shout_util_dict_get(self->meta, name);
+}
+
 int shout_set_public(shout_t *self, unsigned int public)
 {
 	if (!self || (public != 0 && public != 1))
@@ -728,7 +849,8 @@ int shout_set_format(shout_t *self, unsigned int format)
 
 	if (format != SHOUT_FORMAT_OGG
          && format != SHOUT_FORMAT_MP3
-	 && format != SHOUT_FORMAT_WEBM)
+	 && format != SHOUT_FORMAT_WEBM
+	 && format != SHOUT_FORMAT_WEBMAUDIO)
 		return self->error = SHOUTERR_UNSUPPORTED;
 
 	self->format = format;
@@ -754,7 +876,8 @@ int shout_set_protocol(shout_t *self, unsigned int protocol)
 
 	if (protocol != SHOUT_PROTOCOL_HTTP &&
 	    protocol != SHOUT_PROTOCOL_XAUDIOCAST &&
-	    protocol != SHOUT_PROTOCOL_ICY)
+	    protocol != SHOUT_PROTOCOL_ICY &&
+	    protocol != SHOUT_PROTOCOL_ROARAUDIO)
 		return self->error = SHOUTERR_UNSUPPORTED;
 
 	self->protocol = protocol;
@@ -791,146 +914,210 @@ unsigned int shout_get_nonblocking(shout_t *self)
 	return self->nonblocking;
 }
 
-/* -- static function definitions -- */
-
-/* queue data in pages of SHOUT_BUFSIZE bytes */
-static int queue_data(shout_queue_t *queue, const unsigned char *data, size_t len)
+/* TLS functions */
+#ifdef HAVE_OPENSSL
+int shout_set_tls(shout_t *self, int mode)
 {
-	shout_buf_t *buf;
-	size_t plen;
+	if (!self)
+		return SHOUTERR_INSANE;
 
-	if (!len)
-		return SHOUTERR_SUCCESS;
-
-	if (!queue->len) {
-		queue->head = calloc(1, sizeof (shout_buf_t));
-		if (! queue->head)
-			return SHOUTERR_MALLOC;
-	}
-
-	for (buf = queue->head; buf->next; buf = buf->next);
-
-	/* Maybe any added data should be freed if we hit a malloc error?
-	 * Otherwise it'd be impossible to tell where to start requeueing.
-	 * (As if anyone ever tried to recover from a malloc error.) */
-	while (len > 0) {
-		if (buf->len == SHOUT_BUFSIZE) {
-			buf->next = calloc(1, sizeof (shout_buf_t));
-			if (! buf->next)
-				return SHOUTERR_MALLOC;
-			buf->next->prev = buf;
-			buf = buf->next;
-		}
-
-		plen = len > SHOUT_BUFSIZE - buf->len ? SHOUT_BUFSIZE - buf->len : len;
-		memcpy (buf->data + buf->len, data, plen);
-		buf->len += plen;
-		data += plen;
-		len -= plen;
-		queue->len += plen;
-	}
+	if (mode != SHOUT_TLS_DISABLED &&
+	    mode != SHOUT_TLS_AUTO &&
+	    mode != SHOUT_TLS_AUTO_NO_PLAIN &&
+	    mode != SHOUT_TLS_RFC2818 &&
+            mode != SHOUT_TLS_RFC2817)
+		return self->error = SHOUTERR_UNSUPPORTED;
 
+	self->tls_mode = mode;
 	return SHOUTERR_SUCCESS;
 }
-
-static inline int queue_str(shout_t *self, const char *str)
+int shout_get_tls(shout_t *self)
 {
-	return queue_data(&self->wqueue, (const unsigned char*)str, strlen(str));
+	if (!self)
+		return SHOUTERR_INSANE;
+
+	return self->tls_mode;
 }
 
-/* this should be shared with sock_write. Create libicecommon. */
-static int queue_printf(shout_t *self, const char *fmt, ...)
+int shout_set_ca_directory(shout_t *self, const char *directory)
 {
-	char buffer[1024];
-	char *buf;
-	va_list ap, ap_retry;
-	int len;
+	if (!self)
+		return SHOUTERR_INSANE;
 
-	buf = buffer;
+	if (self->state != SHOUT_STATE_UNCONNECTED)
+		return self->error = SHOUTERR_CONNECTED;
 
-	va_start(ap, fmt);
-	va_copy(ap_retry, ap);
+	if (self->ca_directory)
+		free (self->ca_directory);
 
-	len = vsnprintf(buf, sizeof(buffer), fmt, ap);
+	if (!(self->ca_directory = _shout_util_strdup(directory)))
+		return self->error = SHOUTERR_MALLOC;
 
-	self->error = SHOUTERR_SUCCESS;
-	if (len > 0) {
-		if ((size_t)len < sizeof(buffer))
-			queue_data(&self->wqueue, (unsigned char*)buf, len);
-		else {
-			buf = malloc(++len);
-			if (buf) {
-				len = vsnprintf(buf, len, fmt, ap_retry);
-				queue_data(&self->wqueue, (unsigned char*)buf, len);
-				free(buf);
-			} else
-				self->error = SHOUTERR_MALLOC;
-		}
-	}
-
-	va_end(ap_retry);
-	va_end(ap);
-
-	return self->error;
+	return self->error = SHOUTERR_SUCCESS;
 }
 
-static inline void queue_free(shout_queue_t *queue)
+const char *shout_get_ca_directory(shout_t *self)
 {
-	shout_buf_t *prev;
+	if (!self)
+		return NULL;
 
-	while (queue->head) {
-		prev = queue->head;
-		queue->head = queue->head->next;
-		free(prev);
-	}
-	queue->len = 0;
+	return self->ca_directory;
 }
 
+int shout_set_ca_file(shout_t *self, const char *file)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+
+	if (self->state != SHOUT_STATE_UNCONNECTED)
+		return self->error = SHOUTERR_CONNECTED;
+
+	if (self->ca_file)
+		free (self->ca_file);
+
+	if (!(self->ca_file = _shout_util_strdup(file)))
+		return self->error = SHOUTERR_MALLOC;
+
+	return self->error = SHOUTERR_SUCCESS;
+}
+
+const char *shout_get_ca_file(shout_t *self)
+{
+	if (!self)
+		return NULL;
+
+	return self->ca_file;
+}
+
+int shout_set_allowed_ciphers(shout_t *self, const char *ciphers)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+
+	if (self->state != SHOUT_STATE_UNCONNECTED)
+		return self->error = SHOUTERR_CONNECTED;
+
+	if (self->allowed_ciphers)
+		free (self->allowed_ciphers);
+
+	if (!(self->allowed_ciphers = _shout_util_strdup(ciphers)))
+		return self->error = SHOUTERR_MALLOC;
+
+	return self->error = SHOUTERR_SUCCESS;
+}
+
+const char *shout_get_allowed_ciphers(shout_t *self)
+{
+	if (!self)
+		return NULL;
+
+	return self->allowed_ciphers;
+}
+
+int shout_set_client_certificate(shout_t *self, const char *certificate)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+
+	if (self->state != SHOUT_STATE_UNCONNECTED)
+		return self->error = SHOUTERR_CONNECTED;
+
+	if (self->client_certificate)
+		free (self->client_certificate);
+
+	if (!(self->client_certificate = _shout_util_strdup(certificate)))
+		return self->error = SHOUTERR_MALLOC;
+
+	return self->error = SHOUTERR_SUCCESS;
+}
+
+const char *shout_get_client_certificate(shout_t *self)
+{
+	if (!self)
+		return NULL;
+
+	return self->client_certificate;
+}
+#else
+int shout_set_tls(shout_t *self, int mode)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+
+	if (mode == SHOUT_TLS_DISABLED)
+		return SHOUTERR_SUCCESS;
+
+	return self->error = SHOUTERR_UNSUPPORTED;
+}
+int shout_get_tls(shout_t *self)
+{
+	return SHOUT_TLS_DISABLED;
+}
+int shout_set_ca_directory(shout_t *self, const char *directory)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+	return self->error = SHOUTERR_UNSUPPORTED;
+}
+const char *shout_get_ca_directory(shout_t *self)
+{
+	return NULL;
+}
+
+int shout_set_ca_file(shout_t *self, const char *file)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+	return self->error = SHOUTERR_UNSUPPORTED;
+}
+const char *shout_get_ca_file(shout_t *self)
+{
+	return NULL;
+}
+
+int shout_set_allowed_ciphers(shout_t *self, const char *ciphers)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+	return self->error = SHOUTERR_UNSUPPORTED;
+}
+const char *shout_get_allowed_ciphers(shout_t *self)
+{
+	return NULL;
+}
+
+int shout_set_client_certificate(shout_t *self, const char *certificate)
+{
+	if (!self)
+		return SHOUTERR_INSANE;
+	return self->error = SHOUTERR_UNSUPPORTED;
+}
+const char *shout_get_client_certificate(shout_t *self)
+{
+	return NULL;
+}
+#endif
+
+/* -- static function definitions -- */
 static int get_response(shout_t *self)
 {
 	char buf[1024];
-	int rc, blen;
-	char *pc;
-	shout_buf_t *queue;
-	int newlines = 0;
+	int rc;
 
-	rc = sock_read_bytes(self->socket, buf, sizeof(buf));
+	rc = shout_conn_read(self, buf, sizeof(buf));
 
-	if (rc < 0 && sock_recoverable(sock_error()))
+	if (rc < 0 && shout_conn_recoverable(self))
 		return SHOUTERR_BUSY;
 	if (rc <= 0)
 		return SHOUTERR_SOCKET;
 
-	if ((rc = queue_data(&self->rqueue, (unsigned char*)buf, rc)))
+	if ((rc = shout_queue_data(&self->rqueue, (unsigned char*)buf, rc)))
 		return rc;
 
-	/* work from the back looking for \r?\n\r?\n. Anything else means more
-	 * is coming. */
-	for (queue = self->rqueue.head; queue->next; queue = queue->next);
-	pc = (char*)queue->data + queue->len - 1;
-	blen = queue->len;
-	while (blen) {
-		if (*pc == '\n')
-			newlines++;
-		/* we may have to scan the entire queue if we got a response with
-		 * data after the head line (this can happen with eg 401) */
-		else if (*pc != '\r')
-			newlines = 0;
+        if (self->protocol == SHOUT_PROTOCOL_ROARAUDIO)
+            return shout_get_roaraudio_response(self);
 
-		if (newlines == 2)
-			return SHOUTERR_SUCCESS;
-
-		blen--;
-		pc--;
-
-		if (!blen && queue->prev) {
-			queue = queue->prev;
-			pc = (char*)queue->data + queue->len - 1;
-			blen = queue->len;
-		}
-	}
-
-	return SHOUTERR_BUSY;
+	return shout_get_http_response(self);
 }
 
 static int try_connect (shout_t *self)
@@ -939,6 +1126,7 @@ static int try_connect (shout_t *self)
 	int port;
 
 	/* the breaks between cases are omitted intentionally */
+retry:
 	switch (self->state) {
 	case SHOUT_STATE_UNCONNECTED:
 		port = self->port;
@@ -952,9 +1140,7 @@ static int try_connect (shout_t *self)
 		} else {
 			if ((self->socket = sock_connect(self->host, port)) < 0)
 				return self->error = SHOUTERR_NOCONNECT;
-			if ((rc = create_request(self)) != SHOUTERR_SUCCESS)
-				return rc;
-			self->state = SHOUT_STATE_REQ_PENDING;
+			self->state = SHOUT_STATE_CONNECT_PENDING;
 		}
 
 	case SHOUT_STATE_CONNECT_PENDING:
@@ -966,9 +1152,55 @@ static int try_connect (shout_t *self)
 				} else
 					return SHOUTERR_BUSY;
 			}
-			if ((rc = create_request(self)) != SHOUTERR_SUCCESS)
-                                goto failure;
 		}
+		self->state = SHOUT_STATE_TLS_PENDING;
+
+	case SHOUT_STATE_TLS_PENDING:
+#ifdef HAVE_OPENSSL
+		if (self->tls_mode == SHOUT_TLS_DISABLED) {
+			/* nothing to be done */
+		} else if (self->tls_mode == SHOUT_TLS_AUTO || self->tls_mode == SHOUT_TLS_AUTO_NO_PLAIN) {
+			if (self->server_caps & LIBSHOUT_CAP_GOTCAPS) {
+				/* We had a probe allready, otherwise just do nothing to poke the server. */
+				if (self->server_caps & LIBSHOUT_CAP_UPGRADETLS) {
+					self->tls_mode = SHOUT_TLS_RFC2817;
+				} else {
+					if (self->tls_mode == SHOUT_TLS_AUTO_NO_PLAIN)
+						return SHOUTERR_NOTLS;
+					self->tls_mode = SHOUT_TLS_DISABLED;
+				}
+				self->state = SHOUT_STATE_TLS_PENDING;
+				goto retry;
+			}
+		} else if (self->tls_mode == SHOUT_TLS_RFC2818 || self->upgrade_to_tls) {
+			if (!self->tls) {
+				self->tls = shout_tls_new(self, self->socket);
+				if (!self->tls) /* just guessing that it's a malloc error */
+					return SHOUTERR_MALLOC;
+			}
+			if ((rc = shout_tls_try_connect(self->tls)) != SHOUTERR_SUCCESS) {
+				if (rc == SHOUTERR_BUSY)
+					return SHOUTERR_BUSY;
+				goto failure;
+			}
+		} else if (self->tls_mode == SHOUT_TLS_RFC2817) {
+			if ((rc = shout_create_http_request_upgrade(self, "TLS/1.0")) != SHOUTERR_SUCCESS) {
+				if (rc == SHOUTERR_BUSY)
+					return SHOUTERR_BUSY;
+				goto failure;
+			}
+			self->state = SHOUT_STATE_REQ_PENDING;
+			goto retry;
+		} else {
+                        rc = SHOUTERR_INSANE;
+                        goto failure;
+		}
+#endif
+		self->state = SHOUT_STATE_REQ_CREATION;
+
+	case SHOUT_STATE_REQ_CREATION:
+		if ((rc = create_request(self)) != SHOUTERR_SUCCESS)
+			goto failure;
 		self->state = SHOUT_STATE_REQ_PENDING;
 
 	case SHOUT_STATE_REQ_PENDING:
@@ -977,6 +1209,12 @@ static int try_connect (shout_t *self)
 		while (!shout_get_nonblocking(self) && rc == SHOUTERR_BUSY);
                 if (rc == SHOUTERR_BUSY)
                         return rc;
+
+                if (rc == SHOUTERR_SOCKET && self->retry) {
+			self->state = SHOUT_STATE_RECONNECT;
+			goto retry;
+		}
+
 		if (rc != SHOUTERR_SUCCESS)
                         goto failure;
 		self->state = SHOUT_STATE_RESP_PENDING;
@@ -988,28 +1226,60 @@ static int try_connect (shout_t *self)
                 if (rc == SHOUTERR_BUSY)
                         return rc;
 
+                if (rc == SHOUTERR_SOCKET && self->retry) {
+			self->state = SHOUT_STATE_RECONNECT;
+			goto retry;
+#ifdef HAVE_OPENSSL
+		} else if (rc == SHOUTERR_SOCKET && !(self->server_caps & LIBSHOUT_CAP_GOTCAPS) &&
+			   (self->tls_mode == SHOUT_TLS_AUTO || self->tls_mode == SHOUT_TLS_AUTO_NO_PLAIN)) {
+			self->state = SHOUT_STATE_RECONNECT;
+			self->tls_mode = SHOUT_TLS_RFC2818;
+			goto retry;
+#endif
+		}
+
 		if (rc != SHOUTERR_SUCCESS)
                         goto failure;
 
-		if ((rc = parse_response(self)) != SHOUTERR_SUCCESS)
+		if ((rc = parse_response(self)) != SHOUTERR_SUCCESS) {
+			if (rc == SHOUTERR_RETRY)
+				goto retry;
+			if (self->retry) {
+				self->state = SHOUT_STATE_TLS_PENDING;
+				goto retry;
+			}
                         goto failure;
+		}
 
-		if (self->format == SHOUT_FORMAT_OGG) {
+		switch (self->format) {
+		case SHOUT_FORMAT_OGG:
 			if ((rc = self->error = shout_open_ogg(self)) != SHOUTERR_SUCCESS)
                                 goto failure;
-		} else if (self->format == SHOUT_FORMAT_MP3) {
+			break;
+		case SHOUT_FORMAT_MP3:
 			if ((rc = self->error = shout_open_mp3(self)) != SHOUTERR_SUCCESS)
                                 goto failure;
-		} else if (self->format == SHOUT_FORMAT_WEBM) {
+			break;
+		case SHOUT_FORMAT_WEBM:
+		case SHOUT_FORMAT_WEBMAUDIO:
 			if ((rc = self->error = shout_open_webm(self)) != SHOUTERR_SUCCESS)
 				goto failure;
-		} else {
+			break;
+		default:
                         rc = SHOUTERR_INSANE;
                         goto failure;
 		}
 
 	case SHOUT_STATE_CONNECTED:
 		self->state = SHOUT_STATE_CONNECTED;
+	break;
+
+	/* special case, no fallthru to this */
+	case SHOUT_STATE_RECONNECT:
+		sock_close(self->socket);
+		self->state = SHOUT_STATE_UNCONNECTED;
+		goto retry;
+	break;
 	}
 	
 	return SHOUTERR_SUCCESS;
@@ -1027,14 +1297,14 @@ static int try_write (shout_t *self, const void *data_p, size_t len)
 
     /* loop until whole buffer is written (unless it would block) */
     do {
-        ret = sock_write_bytes (self->socket, data + pos, len - pos);
+        ret = shout_conn_write(self, data + pos, len - pos);
         if (ret > 0)
             pos += ret;
     } while (pos < len && ret >= 0);
 
     if (ret < 0)
     {
-        if (sock_recoverable (sock_error()))
+        if (shout_conn_recoverable(self))
         {
             self->error = SHOUTERR_BUSY;
             return pos;
@@ -1046,25 +1316,29 @@ static int try_write (shout_t *self, const void *data_p, size_t len)
     return pos;
 }
 
-/* collect nodes of a queue into a single buffer */
-static int collect_queue(shout_buf_t *queue, char **buf)
+ssize_t shout_conn_read(shout_t *self, void *buf, size_t len)
 {
-	shout_buf_t *node;
-	int pos = 0;
-	int len = 0;
-
-	for (node = queue; node; node = node->next)
-		len += node->len;
-
-	if (!(*buf = malloc(len)))
-		return SHOUTERR_MALLOC;
-
-	for (node = queue; node; node = node->next) {
-		memcpy(*buf + pos, node->data, node->len);
-		pos += node->len;
-	}
-
-	return len;
+#ifdef HAVE_OPENSSL
+	if (self->tls)
+		return shout_tls_read(self->tls, buf, len);
+#endif
+	return sock_read_bytes(self->socket, buf, len);
+}
+ssize_t shout_conn_write(shout_t *self, const void *buf, size_t len)
+{
+#ifdef HAVE_OPENSSL
+	if (self->tls)
+		return shout_tls_write(self->tls, buf, len);
+#endif
+	return sock_write_bytes(self->socket, buf, len);
+}
+int shout_conn_recoverable(shout_t *self)
+{
+#ifdef HAVE_OPENSSL
+	if (self->tls)
+		return shout_tls_recoverable(self->tls);
+#endif
+	return sock_recoverable(sock_error());
 }
 
 static int send_queue(shout_t *self)
@@ -1099,222 +1373,26 @@ static int send_queue(shout_t *self)
 static int create_request(shout_t *self)
 {
 	if (self->protocol == SHOUT_PROTOCOL_HTTP)
-		return create_http_request(self);
+		return shout_create_http_request(self);
 	else if (self->protocol == SHOUT_PROTOCOL_XAUDIOCAST)
-		return create_xaudiocast_request(self);
+		return shout_create_xaudiocast_request(self);
 	else if (self->protocol == SHOUT_PROTOCOL_ICY)
-		return create_icy_request(self);
+		return shout_create_icy_request(self);
+	else if (self->protocol == SHOUT_PROTOCOL_ROARAUDIO)
+		return shout_create_roaraudio_request(self);
 
 	return self->error = SHOUTERR_UNSUPPORTED;
 }
 
-static int create_http_request(shout_t *self)
-{
-	char *auth;
-	char *ai;
-	int ret = SHOUTERR_MALLOC;
-
-	/* this is lazy code that relies on the only error from queue_* being
-	 * SHOUTERR_MALLOC */
-	do {
-		if (queue_printf(self, "SOURCE %s HTTP/1.0\r\n", self->mount))
-			break;
-		if (self->password) {
-			if (! (auth = http_basic_authorization(self)))
-				break;
-			if (queue_str(self, auth)) {
-				free(auth);
-				break;
-			}
-			free(auth);
-		}
-		if (self->useragent && queue_printf(self, "User-Agent: %s\r\n", self->useragent))
-			break;
-		if (self->format == SHOUT_FORMAT_OGG && queue_printf(self, "Content-Type: application/ogg\r\n"))
-			break;
-		if (self->format == SHOUT_FORMAT_MP3 && queue_printf(self, "Content-Type: audio/mpeg\r\n"))
-			break;
-		if (self->format == SHOUT_FORMAT_WEBM && queue_printf(self, "Content-Type: video/webm\r\n"))
-			break;
-		if (queue_printf(self, "ice-name: %s\r\n", self->name ? self->name : "no name"))
-			break;
-		if (queue_printf(self, "ice-public: %d\r\n", self->public))
-			break;
-
-		if (self->url && queue_printf(self, "ice-url: %s\r\n", self->url))
-			break;
-		if (self->genre && queue_printf(self, "ice-genre: %s\r\n", self->genre))
-			break;
-		if ((ai = _shout_util_dict_urlencode(self->audio_info, ';'))) {
-			if (queue_printf(self, "ice-audio-info: %s\r\n", ai)) {
-				free(ai);
-				break;
-			}
-			free(ai);
-		}
-		if (self->description && queue_printf(self, "ice-description: %s\r\n", self->description))
-			break;
-		if (queue_str(self, "\r\n"))
-			break;
-		
-		ret = SHOUTERR_SUCCESS;
-	} while (0);
-
-	return ret;
-}
-
-static char *http_basic_authorization(shout_t *self)
-{
-	char *out, *in;
-	int len;
-
-	if (!self || !self->user || !self->password)
-		return NULL;
-
-	len = strlen(self->user) + strlen(self->password) + 2;
-	if (!(in = malloc(len)))
-		return NULL;
-	sprintf(in, "%s:%s", self->user, self->password);
-	out = _shout_util_base64_encode(in);
-	free(in);
-
-	len = strlen(out) + 24;
-	if (!(in = malloc(len))) {
-		free(out);
-		return NULL;
-	}
-	sprintf(in, "Authorization: Basic %s\r\n", out);
-	free(out);
-	
-	return in;
-}
-
 static int parse_response(shout_t *self)
 {
 	if (self->protocol == SHOUT_PROTOCOL_HTTP)
-		return parse_http_response(self);
+		return shout_parse_http_response(self);
 	else if (self->protocol == SHOUT_PROTOCOL_XAUDIOCAST ||
 		 self->protocol == SHOUT_PROTOCOL_ICY)
-		return parse_xaudiocast_response(self);
+		return shout_parse_xaudiocast_response(self);
+	else if (self->protocol == SHOUT_PROTOCOL_ROARAUDIO)
+		return shout_parse_roaraudio_response(self);
 
 	return self->error = SHOUTERR_UNSUPPORTED;
 }
-
-static int parse_http_response(shout_t *self)
-{
-	http_parser_t *parser;
-	char *header = NULL;
-	int hlen = 0;
-	int code;
-	char *retcode;
-#if 0
-	char *realm;
-#endif
-
-	/* all this copying! */
-	hlen = collect_queue(self->rqueue.head, &header);
-	if (hlen <= 0)
-		return SHOUTERR_MALLOC;
-	queue_free(&self->rqueue);
-
-	parser = httpp_create_parser();
-	httpp_initialize(parser, NULL);
-	if (httpp_parse_response(parser, header, hlen, self->mount)) {
-		retcode = httpp_getvar(parser, HTTPP_VAR_ERROR_CODE);
-		code = atoi(retcode);
-		if(code >= 200 && code < 300) {
-			httpp_destroy(parser);
-			free (header);
-			return SHOUTERR_SUCCESS;
-		}
-	}
-
-	free(header);
-	httpp_destroy(parser);
-	return self->error = SHOUTERR_NOLOGIN;
-}
-
-static int create_xaudiocast_request(shout_t *self)
-{
-	const char *bitrate;
-	int ret;
-
-	bitrate = shout_get_audio_info(self, SHOUT_AI_BITRATE);
-	if (!bitrate)
-		bitrate = "0";
-
-	ret = SHOUTERR_MALLOC;
-	do {
-		if (queue_printf(self, "SOURCE %s %s\n", self->password, self->mount))
-			break;
-		if (queue_printf(self, "x-audiocast-name: %s\n", self->name ? self->name : "unnamed"))
-			break;
-		if (queue_printf(self, "x-audiocast-url: %s\n", self->url ? self->url : "http://www.icecast.org/"))
-			break;
-		if (queue_printf(self, "x-audiocast-genre: %s\n", self->genre ? self->genre : "icecast"))
-			break;
-		if (queue_printf(self, "x-audiocast-bitrate: %s\n", bitrate))
-			break;
-		if (queue_printf(self, "x-audiocast-public: %i\n", self->public))
-			break;
-		if (queue_printf(self, "x-audiocast-description: %s\n", self->description ? self->description : "Broadcasting with the icecast streaming media server!"))
-			break;
-		if (self->dumpfile && queue_printf(self, "x-audiocast-dumpfile: %s\n", self->dumpfile))
-			break;
-		if (queue_str(self, "\n"))
-			break;
-
-		ret = SHOUTERR_SUCCESS;
-	} while (0);
-		
-	return ret;
-}
-
-static int parse_xaudiocast_response(shout_t *self)
-{
-	char *response;
-
-	if (collect_queue(self->rqueue.head, &response) <= 0)
-		return SHOUTERR_MALLOC;
-	queue_free(&self->rqueue);
-
-	if (!strstr(response, "OK")) {
-		free(response);
-		return SHOUTERR_NOLOGIN;
-	}
-	free(response);
-
-	return SHOUTERR_SUCCESS;
-}
-
-static int create_icy_request(shout_t *self)
-{
-	const char *bitrate;
-	int ret;
-
-	bitrate = shout_get_audio_info(self, SHOUT_AI_BITRATE);
-	if (!bitrate)
-		bitrate = "0";
-
-	ret = SHOUTERR_MALLOC;
-	do {
-		if (queue_printf(self, "%s\n", self->password))
-			break;
-		if (queue_printf(self, "icy-name:%s\n", self->name ? self->name : "unnamed"))
-			break;
-		if (queue_printf(self, "icy-url:%s\n", self->url ? self->url : "http://www.icecast.org/"))
-			break;
-		if (queue_str(self, "icy-irc:\nicy-aim:\nicy-icq:\n"))
-			break;
-		if (queue_printf(self, "icy-pub:%i\n", self->public))
-			break;
-		if (queue_printf(self, "icy-genre:%s\n", self->genre ? self->genre : "icecast"))
-			break;
-		if (queue_printf(self, "icy-br:%s\n\n", bitrate))
-			break;
-
-		ret = SHOUTERR_SUCCESS;
-	} while (0);
-
-	return ret;
-}
